from graphite.util import jsonResponse, HttpResponse, HttpError
from graphite.storage import STORE, extractForwardHeaders


def _requestContext(request, queryParams):
    return {
        'forwardHeaders': extractForwardHeaders(request),
        'localOnly': queryParams.get('local') == '1',
    }


@jsonResponse
def tagSeries(request, queryParams):
    if request.method != 'POST':
        return HttpResponse(status=405)

    path = queryParams.get('path')
    if not path:
        raise HttpError('no path specified', status=400)

    return STORE.tagdb.tag_series(path, requestContext=_requestContext(request, queryParams))


@jsonResponse
def tagMultiSeries(request, queryParams):
    if request.method != 'POST':
        return HttpResponse(status=405)

    paths = []
    # Normal format: ?path=name;tag1=value1;tag2=value2&path=name;tag1=value2;tag2=value2
    if len(queryParams.getlist('path')) > 0:
        paths = queryParams.getlist('path')
    # Rails/PHP/jQuery common practice format: ?path[]=...&path[]=...
    elif len(queryParams.getlist('path[]')) > 0:
        paths = queryParams.getlist('path[]')
    else:
        raise HttpError('no paths specified',status=400)

    return STORE.tagdb.tag_multi_series(paths, requestContext=_requestContext(request, queryParams))


@jsonResponse
def delSeries(request, queryParams):
    if request.method != 'POST':
        return HttpResponse(status=405)

    paths = []
    # Normal format: ?path=name;tag1=value1;tag2=value2&path=name;tag1=value2;tag2=value2
    if len(queryParams.getlist('path')) > 0:
        paths = queryParams.getlist('path')
    # Rails/PHP/jQuery common practice format: ?path[]=...&path[]=...
    elif len(queryParams.getlist('path[]')) > 0:
        paths = queryParams.getlist('path[]')
    else:
        raise HttpError('no path specified', status=400)

    return STORE.tagdb.del_multi_series(paths, requestContext=_requestContext(request, queryParams))


@jsonResponse
def findSeries(request, queryParams):
    if request.method not in ['GET', 'POST']:
        return HttpResponse(status=405)

    exprs = []
    # Normal format: ?expr=tag1=value1&expr=tag2=value2
    if len(queryParams.getlist('expr')) > 0:
        exprs = queryParams.getlist('expr')
    # Rails/PHP/jQuery common practice format: ?expr[]=tag1=value1&expr[]=tag2=value2
    elif len(queryParams.getlist('expr[]')) > 0:
        exprs = queryParams.getlist('expr[]')

    if not exprs:
        raise HttpError('no tag expressions specified', status=400)

    return STORE.tagdb.find_series(exprs, requestContext=_requestContext(request, queryParams))


@jsonResponse
def tagList(request, queryParams):
    if request.method != 'GET':
        return HttpResponse(status=405)

    return STORE.tagdb.list_tags(
        tagFilter=request.GET.get('filter'),
        limit=request.GET.get('limit'),
        requestContext=_requestContext(request, queryParams),
    )


@jsonResponse
def tagDetails(request, queryParams, tag):
    if request.method != 'GET':
        return HttpResponse(status=405)

    return STORE.tagdb.get_tag(
        tag,
        valueFilter=queryParams.get('filter'),
        limit=queryParams.get('limit'),
        requestContext=_requestContext(request, queryParams),
    )


@jsonResponse
def autoCompleteTags(request, queryParams):
    if request.method not in ['GET', 'POST']:
        return HttpResponse(status=405)

    exprs = []
    # Normal format: ?expr=tag1=value1&expr=tag2=value2
    if len(queryParams.getlist('expr')) > 0:
        exprs = queryParams.getlist('expr')
    # Rails/PHP/jQuery common practice format: ?expr[]=tag1=value1&expr[]=tag2=value2
    elif len(queryParams.getlist('expr[]')) > 0:
        exprs = queryParams.getlist('expr[]')

    return STORE.tagdb_auto_complete_tags(
        exprs,
        tagPrefix=queryParams.get('tagPrefix'),
        limit=queryParams.get('limit'),
        requestContext=_requestContext(request, queryParams)
    )


@jsonResponse
def autoCompleteValues(request, queryParams):
    if request.method not in ['GET', 'POST']:
        return HttpResponse(status=405)

    exprs = []
    # Normal format: ?expr=tag1=value1&expr=tag2=value2
    if len(queryParams.getlist('expr')) > 0:
        exprs = queryParams.getlist('expr')
    # Rails/PHP/jQuery common practice format: ?expr[]=tag1=value1&expr[]=tag2=value2
    elif len(queryParams.getlist('expr[]')) > 0:
        exprs = queryParams.getlist('expr[]')

    tag = queryParams.get('tag')
    if not tag:
        raise HttpError('no tag specified', status=400)

    return STORE.tagdb_auto_complete_values(
        exprs,
        tag,
        valuePrefix=queryParams.get('valuePrefix'),
        limit=queryParams.get('limit'),
        requestContext=_requestContext(request, queryParams)
    )



import datetime
import six

try:
    from django.contrib.sites.requests import RequestSite
except ImportError:  # Django < 1.9
    from django.contrib.sites.models import RequestSite

from django.core.exceptions import ObjectDoesNotExist
from django.core.serializers.json import DjangoJSONEncoder
from django.forms.models import model_to_dict
from django.shortcuts import render, get_object_or_404
from django.utils.timezone import now

from graphite.util import json, epoch, epoch_to_dt, jsonResponse, HttpError, HttpResponse
from graphite.events.models import Event
from graphite.render.attime import parseATTime


class EventEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return epoch(obj)
        return json.JSONEncoder.default(self, obj)


def view_events(request):
    if request.method == 'GET':
        context = {'events': fetch(request),
                   'site': RequestSite(request),
                   'protocol': 'https' if request.is_secure() else 'http'}
        return render(request, 'events.html', context)
    else:
        return post_event(request)


@jsonResponse(encoder=DjangoJSONEncoder)
def jsonDetail(request, queryParams, event_id):
    try:
        e = Event.objects.get(id=event_id)
        e.tags = e.tags.split()
        return model_to_dict(e)
    except ObjectDoesNotExist:
        raise HttpError('Event matching query does not exist', status=404)


def detail(request, event_id):
    if request.META.get('HTTP_ACCEPT') == 'application/json':
        return jsonDetail(request, event_id)

    e = get_object_or_404(Event, pk=event_id)
    context = {'event': e}
    return render(request, 'event.html', context)


def post_event(request):
    if request.method == 'POST':
        event = json.loads(request.body)
        assert isinstance(event, dict)

        tags = event.get('tags')
        if tags is not None:
            if isinstance(tags, list):
                tags = ' '.join(tags)
            elif not isinstance(tags, six.string_types):
                return HttpResponse(
                    json.dumps({'error': '"tags" must be an array or space-separated string'}),
                    status=400)
        else:
            tags = None
        if 'when' in event:
            when = epoch_to_dt(event['when'])
        else:
            when = now()

        Event.objects.create(
            what=event.get('what'),
            tags=tags,
            when=when,
            data=event.get('data', ''),
        )

        return HttpResponse(status=200)
    else:
        return HttpResponse(status=405)


def get_data(request):
    query_params = request.GET.copy()
    query_params.update(request.POST)

    if 'jsonp' in query_params:
        response = HttpResponse(
          "%s(%s)" % (query_params.get('jsonp'),
              json.dumps(fetch(request), cls=EventEncoder)),
          content_type='text/javascript')
    else:
        response = HttpResponse(
            json.dumps(fetch(request), cls=EventEncoder),
            content_type='application/json')
    return response


def fetch(request):
    if request.GET.get('from') is not None:
        time_from = parseATTime(request.GET['from'])
    else:
        time_from = epoch_to_dt(0)

    if request.GET.get('until') is not None:
        time_until = parseATTime(request.GET['until'])
    else:
        time_until = now()

    set_operation = request.GET.get('set')

    tags = request.GET.get('tags')
    if tags is not None:
        tags = request.GET.get('tags').split(' ')

    result = []
    for x in Event.find_events(time_from, time_until, tags=tags, set_operation=set_operation):

        # django-tagging's with_intersection() returns matches with unknown tags
        # this is a workaround to ensure we only return positive matches
        if set_operation == 'intersection':
            if len(set(tags) & set(x.as_dict()['tags'])) == len(tags):
                result.append(x.as_dict())
        else:
            result.append(x.as_dict())
    return result



from django.core.mail import EmailMessage


def send_graph_email(subject, sender, recipients, attachments=None, body=None):
    """
    :param str sender: sender's email address
    :param list recipients: list of recipient emails
    :param list attachments: list of triples of the form:
        (filename, content, mimetype). See the django docs
        https://docs.djangoproject.com/en/1.3/topics/email/#django.core.mail.EmailMessage
    """
    attachments = attachments or []
    msg = EmailMessage(subject=subject,
                       from_email=sender,
                       to=recipients,
                       body=body,
                       attachments=attachments)
    msg.send()



import six

from graphite.errors import InputParameterError
from graphite.render.attime import parseTimeOffset
from graphite.logger import log
from graphite.functions.aggfuncs import aggFuncs, aggFuncAliases
from graphite.render.datalib import TimeSeries


class ParamTypes(object):
    pass


class ParamType(object):
    options = []

    def __init__(self, name, validator=None):
        self.name = name
        self.validator = validator

    @classmethod
    def register(cls, name, *args):
        setattr(ParamTypes, name, cls(name, *args))

    def isValid(self, value):
        if self.validator is None:
            # if there's no validator for the type we assume True
            return value

        return self.validator(value)


def validateBoolean(value):
    if isinstance(value, six.string_types):
        if value.lower() == 'true':
            return True
        if value.lower() == 'false':
            return False
        raise ValueError('Invalid boolean value: {value}'.format(value=repr(value)))

    if type(value) in [int, float]:
        if value == 0:
            return False
        if value == 1:
            return True
        raise ValueError('Invalid boolean value: {value}'.format(value=repr(value)))

    return bool(value)


def validateFloat(value):
    return float(value)


def validateInteger(value):
    # prevent that float values get converted to int, because an
    # error is better than silently falsifying the result
    if type(value) is float:
        raise ValueError('Not a valid integer value: {value}'.format(value=repr(value)))

    return int(value)


def validateIntOrInf(value):
    try:
        return validateInteger(value)
    except (TypeError, ValueError):
        pass

    try:
        inf = float('inf')
        if float(value) == inf:
            return inf
    except (TypeError, ValueError, OverflowError):
        pass

    raise ValueError('Not a valid integer nor float value: {value}'.format(value=repr(value)))


def validateInterval(value):
    try:
        parseTimeOffset(value)
    except (IndexError, KeyError, TypeError, ValueError) as e:
        raise ValueError('Invalid interval value: {value}: {e}'.format(value=repr(value), e=str(e)))
    return value


def validateSeriesList(value):
    if not isinstance(value, list):
        raise ValueError('Invalid value type, it is not a list: {value}'.format(value=repr(value)))

    for series in value:
        if not isinstance(series, TimeSeries):
            raise ValueError('Invalid type "{type}", should be TimeSeries'.format(type=type(series)))

    return value


def validateSeriesLists(value):
    if not isinstance(value, list):
        raise ValueError('Invalid value type, it is not a list: {value}'.format(value=repr(value)))

    for entry in value:
        validateSeriesList(entry)

    return value


ParamType.register('boolean', validateBoolean)
ParamType.register('date')
ParamType.register('float', validateFloat)
ParamType.register('integer', validateInteger)
ParamType.register('interval', validateInterval)
ParamType.register('intOrInterval')
ParamType.register('intOrInf', validateIntOrInf)
ParamType.register('node', validateInteger)
ParamType.register('nodeOrTag')
ParamType.register('series')
ParamType.register('seriesList', validateSeriesList)
ParamType.register('seriesLists', validateSeriesLists)
ParamType.register('string')
ParamType.register('tag')

# special type that accepts everything
ParamType.register('any')


class ParamTypeAggFunc(ParamType):

    def __init__(self, name, validator=None):
        if validator is None:
            validator = self.validateAggFuncs

        super(ParamTypeAggFunc, self).__init__(name=name, validator=validator)
        self.options = self.getValidAggFuncs()

    @classmethod
    def getValidAggFuncs(cls):
        return list(aggFuncs.keys()) + list(aggFuncAliases.keys())

    @classmethod
    def getDeprecatedAggFuncs(cls):
        return [name + 'Series' for name in cls.getValidAggFuncs()]

    @classmethod
    def getAllValidAggFuncs(cls):
        return cls.getValidAggFuncs() + cls.getDeprecatedAggFuncs()

    def validateAggFuncs(self, value):
        if value in self.getValidAggFuncs():
            return value

        if value in self.getDeprecatedAggFuncs():
            log.warning('Deprecated aggregation function: {value}'.format(value=repr(value)))
            return value

        raise ValueError('Invalid aggregation function: {value}'.format(value=repr(value)))


ParamTypeAggFunc.register('aggFunc')


class ParamTypeAggOrSeriesFunc(ParamTypeAggFunc):
    options = []

    def __init__(self, name, validator=None):
        if validator is None:
            validator = self.validateAggOrSeriesFuncs
        super(ParamTypeAggOrSeriesFunc, self).__init__(name=name, validator=validator)

    def setSeriesFuncs(self, funcs):
        # check for each of the series functions whether they have an 'aggregator'
        # property being set to 'True'. If so we consider them valid aggregators.
        for name, func in funcs.items():
            if getattr(func, 'aggregator', False) is not True:
                continue

            self.options.append(name)

    def validateAggOrSeriesFuncs(self, value):
        try:
            return self.validateAggFuncs(value)
        except ValueError:
            pass

        if value in self.options:
            return value

        return False


ParamTypeAggOrSeriesFunc.register('aggOrSeriesFunc')


class Param(object):
    __slots__ = ('name', 'type', 'required', 'default', 'multiple', '_options', 'suggestions')

    def __init__(self, name, paramtype, required=False, default=None, multiple=False, options=None, suggestions=None):
        self.name = name
        if not isinstance(paramtype, ParamType):
            raise Exception('Invalid type %s for parameter %s' % (paramtype, name))
        self.type = paramtype
        self.required = bool(required)
        self.default = default
        self.multiple = bool(multiple)
        if options is None:
            options = []
        self._options = options
        self.suggestions = suggestions

    @property
    def options(self):
        options = list(set(getattr(self, '_options', []) + getattr(self.type, 'options', [])))
        options.sort(key=str)
        return options

    def toJSON(self):
        jsonVal = {
            'name': self.name,
            'type': self.type.name,
        }
        if self.required:
            jsonVal['required'] = True
        if self.default is not None:
            jsonVal['default'] = self.default
        if self.multiple:
            jsonVal['multiple'] = True
        if self.options:
            jsonVal['options'] = self.options
        if self.suggestions:
            jsonVal['suggestions'] = self.suggestions
        return jsonVal

    def validateValue(self, value, func):
        # if value isn't specified and there's a default then the default will be used,
        # we don't need to validate the default value because we trust that it is valid
        if value is None and self.default is not None:
            return value

        # None is ok for optional params
        if not self.required and value is None:
            return value

        # parameter is restricted to a defined set of values, but value is not in it
        if self.options and value not in self.options:
            raise InputParameterError(
                'Invalid option specified for function "{func}" parameter "{param}": {value}'.format(
                    func=func, param=self.name, value=repr(value)))

        try:
            return self.type.isValid(value)
        except Exception:
            raise InputParameterError(
                'Invalid "{type}" value specified for function "{func}" parameter "{param}": {value}'.format(
                    type=self.type.name, func=func, param=self.name, value=repr(value)))


def validateParams(func, params, args, kwargs):
    valid_args = []
    valid_kwargs = {}

    # total number of args + kwargs might be larger than number of params if
    # the last param allows to be specified multiple times
    if len(args) + len(kwargs) > len(params):
        if not params[-1].multiple:
            raise InputParameterError(
                'Too many parameters specified for function "{func}"'.format(func=func))

        # if args has more values than params and the last param allows multiple values,
        # then we're going to validate all paramas from the args value list
        args_params = params
        kwargs_params = []
    else:
        # take the first len(args) params and use them to validate the args values,
        # use the remaining params to validate the kwargs values
        args_params = params[:len(args)]
        kwargs_params = params[len(args):]

    # validate the args
    for (i, arg) in enumerate(args):
        if i >= len(args_params):
            # last parameter must be allowing multiple,
            # so we're using it to validate this arg
            param = args_params[-1]
        else:
            param = args_params[i]

        valid_args.append(param.validateValue(arg, func))

    # validate the kwargs
    for param in kwargs_params:
        value = kwargs.get(param.name, None)
        if value is None:
            if param.required:
                raise InputParameterError(
                    'Missing required parameter "{param}" for function "{func}"'
                    .format(param=param.name, func=func))
            continue

        valid_kwargs[param.name] = param.validateValue(value, func)

    if len(kwargs) > len(valid_kwargs):
        unexpected_keys = []
        for name in kwargs.keys():
            if name not in valid_kwargs:
                unexpected_keys.append(name)
        raise InputParameterError(
            'Unexpected key word arguments: {keys}'.format(
                keys=', '.join(
                    key
                    for key in kwargs.keys()
                    if key not in valid_kwargs.keys()
                )))

    return (valid_args, valid_kwargs)



import functools

from graphite.intervals import IntervalSet
from graphite.logger import log
from graphite.readers.utils import BaseReader


class MultiReader(BaseReader):
    __slots__ = ('nodes',)

    def __init__(self, nodes):
        self.nodes = nodes

    def get_intervals(self):
        interval_sets = []
        for node in self.nodes:
            interval_sets.extend(node.intervals.intervals)
        return IntervalSet(sorted(interval_sets))

    def fetch(self, startTime, endTime, now=None, requestContext=None):
        # Start the fetch on each node
        fetches = []

        for n in self.nodes:
            try:
                fetches.append(n.fetch(startTime, endTime, now, requestContext))
            except BaseException:
                log.exception("Failed to initiate subfetch for %s" % str(n))

        results = [
          r for r in fetches
          if r is not None
        ]
        if not results:
            raise Exception("All sub-fetches failed")

        return functools.reduce(self.merge, results)

    @staticmethod
    def merge(results1, results2):
        # Ensure results1 is finer than results2
        if results1[0][2] > results2[0][2]:
            results1, results2 = results2, results1

        time_info1, values1 = results1
        time_info2, values2 = results2
        start1, end1, step1 = time_info1
        start2, end2, step2 = time_info2

        step = step1                # finest step
        start = min(start1, start2)  # earliest start
        end = max(end1, end2)      # latest end
        time_info = (start, end, step)
        values = []

        t = start
        while t < end:
            # Look for the finer precision value first if available
            i1 = (t - start1) // step1

            if len(values1) > i1:
                v1 = values1[i1]
            else:
                v1 = None

            if v1 is None:
                i2 = (t - start2) // step2

                if len(values2) > i2:
                    v2 = values2[i2]
                else:
                    v2 = None

                values.append(v2)
            else:
                values.append(v1)

            t += step

        return (time_info, values)
